#This is a simple function that numerically finds a solution to an equation of the form A=fn(x), on the interval [upper, lower]. This might not be the only solution, which is why I make sure to use this function only in cases where uniqueness follows from other results.

function bisect_root(fn, lower, upper, lhs)
  x0 = lower
  x2 = upper
  x1 = (x0+x2)/2

    y0 = fn(x0)-lhs
    y1 = fn(x1)-lhs
    y2 = fn(x2)-lhs

  while x0 < x1 && x1 < x2
    if sign(y0) == sign(y1)
      x0, x2 = x1, x2
      y0, y2 = y1, y2
    else
      x0, x2 = x0, x1
      y0, y2 = y0, y1
    end

    x1 = (x0+x2)/2
        y1 = fn(x1)-lhs
  end
x1
end

function bisect_rootalt(fn, lower, upper, lhs)
  x0 = lower
  x2 = upper
  x1 = (x0+x2)/2

    y0 = 0-lhs
    y1 = 1-lhs
    y2 = fn(x2)-lhs

  while x0 < x1 && x1 < x2
    if sign(y0) == sign(y1)
      x0, x2 = x1, x2
      y0, y2 = y1, y2
    else
      x0, x2 = x0, x1
      y0, y2 = y0, y1
    end

    x1 = (x0+x2)/2
        y1 = fn(x1)-lhs
  end
x1
end

#This function calculates map T, given vc, uc and some function  z:e (equivalent of d in the proof of Theorem 1).

function Tmapf(psi::Array{Float64, 1}, z::Array{Float64, 1}, Cuex::Function, Cvex::Function, pi1uex::Function, pi2vex::Function, R2, R1, vc, uc, M1, M2)
    step=(1-vc)/(length(psi)-1)
    test=psi[1]
    Cevd=Cvex.(psi, z)
    Ceud=Cuex.(psi, z)
    Cevd[end]=Cevd[end-1] #This is done because Cvex(dot, v) is not continuous in u. This is not a problem for the proof, as this happens with measure 0, but here it happens with a positive measure and thus does pose a problem. An alternative fix would be to always keep the condition on M's precision smaller than grid (which is approximating the fact that this discontinuity happens with measure 0)
    Ceud[end]=Ceud[end-1]
    M2cum=Array{Float64}(undef,  length(psi))
    M2cum[1]=0
    M1cum=Array{Float64}(undef,  length(psi))
    M1cum[1]=0
    for i=2:length(psi)
        M2cum[i]=M2cum[i-1]+(Cevd[i-1]+Cevd[i])*(step/2) #numerical appoximation of an integral, trapezoid method
        M1cum[i]=M1cum[i-1]+(Ceud[i-1]+Ceud[i])*((psi[i]-psi[i-1])/2)
end
    nom=Array{Float64}(undef,  length(z)) #numerator of map T
    for i=1:length(z)
        nom[i]=pi2vex(z[i], (R2-M2+M2cum[i])/R2)
end
    denom=Array{Float64}(undef,  length(psi)) #denumerator of map T
    for i=1:length(psi)
        denom[i]=pi1uex(psi[i], (R1-M1+M1cum[i])/R1) #numerical appoximation of an integral, trapezoid method
end
    inside=nom./denom
    f=Array{Float64}(undef,  length(psi))
    f[1]=uc
    for i=2:length(psi)
        f[i]=f[i-1]+(inside[i-1]+inside[i])*(step/2) #numerical appoximation of an integral, trapezoid method
end
    return f, M2cum, M1cum, nom, denom
end

#this is the main function that finds the stable matching, by finding the fixed point of map T (see the proof of Theorem 1; the map itself is defined in tmapfunc.jl above) given uc and vc; and then looks for such uc and vc that the mass of agents is equal to the mass of firms.
function equilibriumbase(Cex::Function, Cuex::Function, Cvex::Function, pi1uex::Function, pi2vex::Function, R2, R1, grid, no_step, no_step1, vc::Number, uc::Number, vcmax::Number, vcdash::Number)
        r1=1
        if R1+R2==1
            adash=vcdash
amax=vcmax
            if vc>0
                a=vc
            else
                a=-uc
            end
        end

        #The next line uses the EUler's method to find the fixed point of map T (proof of Theorem 1) for the inition vc, uc
        M2cum, psivecins, z, M1cum, vc, uc=eulermethod(pi1uex::Function, pi2vex::Function, Cvex::Function, Cuex::Function, R2::Number, R1::Number, grid::Number, vc::Number, uc::Number)
        Mass=M2cum[end]
        #The following uses the bisection method to find the equilibrium pair of critical values
        for i=1:no_step1

        # The next loop allows for three possibilities. If the mass of manufacturing workers is very close to the mass of firms, we are done.
        #If the mass of workers is somewhat close to the mass of firms, we enter an interior function, that does the same as the third possibility, but iterates further over the outomces of the Euler method, in order to approximate the solution better.
        #Finally, if we are still far away from the solution, it start iterating over (uc, vc).
        #The idea is to approach the solution using the fast Euler method, but once sufficiently close, start using the more accurate contraction method.
        #The first possibility is left so that for the symmetric equilibria, where the solution is reached immediately, the algorith does not enter into the interior function.
            if (Mass-R2)^2<(0.00001)^2#(0.0000001)^2
                break

            elseif (i>1 && (Mass-R2)^2<(0.4)^2)
                psivecins, M2cum, M1cum=stableint(C, Cuex, Cvex, pi1uex, pi2vex, psivecins, z, M2cum, M1cum, R2, R1, grid, 100, 100, vc, uc, vcmax, vcdash)
        break

        else
            if R1+R2<1
        if Mass>R2 #If not, we change vc and uc. If the mass of agents was too large, we increase vc, otherwise we decrease it.
        vcdash, vcmax=vc, vcmax
        else
        vcdash, vcmax=vcdash, vc
        end
        vc=(vcdash+vcmax)/2
        Cvc(u)=Cex(u, vc)
        uc=bisect_rootalt(Cvc, 0, 1, 1-R1-R2)
    else
        if Mass>R2
            adash, amax=a, amax
        else
            adash, amax=adash, a
        end
    a=(adash+amax)/2
                if a<0
                    vc=0
                    uc=-a
                else
                    vc=a
                    uc=0
                end
            end
        r1=i+1
        #Finally, we calcualte the fixed point of map T (again, using Euler's method) for the new pair uc, vc
        M2cum, psivecins, z, M1cum, vc, uc=eulermethod(pi1uex::Function, pi2vex::Function, Cvex::Function, Cuex::Function, R2::Number, R1::Number, grid::Number, vc::Number, uc::Number)
        Mass=M2cum[end]
        end
    end
        if r1>=no_step1 #This let's us know if the above loop did not converge
                    print("no convergence")
                end
        return    psivecins, z, M2cum, M1cum, uc, vc, r1
end

#This is a function that approximates the solution of a differential equation using Euler's method

function eulermethod(pi1uex::Function, pi2vex::Function, Cvex::Function, Cuex::Function, R2::Number, R1::Number, grid::Number, vc::Number, uc::Number)
#First, define a few arrays that will be used to store the results
    z=Array{Float64}(undef,  grid+1) #the array of argument
    M1cum=Array{Float64}(undef,  grid+1) #the array of values of the cumulative mass function for manufacturing
    M2cum=Array{Float64}(undef,  grid+1) #the array of values of the cumulative mass function for services
    psivecins=Array{Float64}(undef,  grid+1) #the array of values of the separation function
#The next fixe lines create the vector of arguments
    step=(1-vc)/grid
    z[1]=vc
    for i=2:grid+1
    z[i]=step*(i-1)+z[1] #vector of arguments
    end
    #The next 8 lines set the initial values
    psivecins[1]=uc
    vprev=z[1]
    M1cum[1]=0
    M2cum[1]=0
    j=1
    psiprev=psivecins[1]
    M2=0
    M1=0
    #The next loop implemets the Euler's method to approximate the fixed point of map T (proof of Theorem 1)
    for i=2:(length(z))
#First, we evalueate the derivative of the separation function, using the values of the argument from the previous step
    up=pi2vex(vprev, M2/R2)
    down=pi1uex(psiprev, M1/R1)
        psiprime=pi2vex(vprev, M2/R2)/pi1uex(psiprev, M1/R1)
        # We use this to update the value of the separation function
    psi=psiprev+psiprime*step
    #and then update the balue of the arguemnt
    v=vprev+step
    #using those update values, we deploy the trapezoid method to update the values of the mass of workers in each sector
    M2=M2+(Cvex(psiprev, vprev)+Cvex(psi, v))*step/2
        M1=M1+(Cuex(psiprev, vprev)+Cuex(psi, v))*psiprime*step/2
        #The new values a stored within the previously defined matrices
    M1cum[i]=M1
    M2cum[i]=M2
        psivecins[i]=psi
        #and, finally, we update the "old" values of psi and v
        vprev=v
        psiprev=psi
    end
    return M2cum, psivecins, z, M1cum, vc, uc
end


#This is an internal function used to improve the precision of the calculations
#The idea is the same as in the external function but here the Euler method's solution is further iterated, to make sure the actual fixed point has been found
function stableint(Cex::Function, Cuex::Function, Cvex::Function, pi1uex::Function, pi2vex::Function, psivecins::Array, z::Array, M2cum::Array, M1cum::Array, R2, R1, grid, no_step, no_step1, vc, uc, vcmax, vcdash)
    r=1
    r1=1
    if R1+R2==1
        adash=vcdash
amax=vcmax
        if vc>0
            a=vc
        else
            a=-uc
        end
    end
    #The following loop keeps iterating over function psi, until the former and latter iterations become sufficently close.
    for i=1:no_step
    r=i

         old=psivecins
         psivecins, M2cum, M1cum = Tmapf(psivecins, z, Cuex, Cvex, pi1uex, pi2vex, R2, R1, vc, uc, R1, R2)
         dfr=(psivecins-old).^2
         maxdfr=maximum(dfr)
         #print("$maxdfr $r \n")
    if maximum(dfr)<(0.000001)^2
     break
    end
    end
     if r>=no_step #This let's us know if the above loop did not converge
                 print("no convergence")
             end
    Mass=M2cum[end]
    for i=1:no_step1
        #The next loop iterates over (uc, vc) until the mass of manufacturing workers becomes sufficiently close to the emasure of firms.
    if (Mass-R2)^2<(0.00001)^2
     break
    else
        if R1+R2<1
    if Mass>R2 #If not, we change vc and uc. If the mass of agents was too large, we increase vc, otherwise we decrease it.
    vcdash, vcmax=vc, vcmax
    else
    vcdash, vcmax=vcdash, vc
    end
    vc=(vcdash+vcmax)/2
    Cvc(u)=Cex(u, vc)
    uc=bisect_rootalt(Cvc, 0, 1, 1-R1-R2)
else
    if Mass>R2
        adash, amax=a, amax
    else
        adash, amax=adash, a
    end
a=(adash+amax)/2
            if a<0
                vc=0
                uc=-a
            else
                vc=a
                uc=0
            end
        end
        r1=i+1
        #print("$r1 ")
        #print("$vc $uc $adash $amax ")
        #print("$Mass ")
                M2cum, psivecins, z, M1cum, vc, uc=eulermethod(pi1uex::Function, pi2vex::Function, Cvex::Function, Cuex::Function, R2::Number, R1::Number, grid::Number, vc::Number, uc::Number)
#Next: same as above, iteration over psi.
    for i=1:no_step
                 r=i
         old=psivecins
         psivecins, M2cum, M1cum = Tmapf(psivecins, z, Cuex, Cvex, pi1uex, pi2vex, R2, R1, vc, uc, R1, R2)
         dfr=(psivecins-old).^2
         maxdfr=maximum(dfr)
    if maximum(dfr)<(0.000001)^2
     break
    end
    end
             if r>=no_step #This let's us know if the above loop did not converge
                 print("no convergence")
             end
    Mass=M2cum[end]
    end
    end
     if r1>=no_step1 #This let's us know if the above loop did not converge
                 print("no convergence")
             end
     return    psivecins, M2cum, M1cum, uc, vc, r1
end


#this is a function that given f: z->d plus the vector of arguments of f's inverse,  returns the values of that inverse. It's defined a few times to allow foer different input types.
function inverse(u::Array{Float64, 1}, d::Array{Float64, 1}, z::Array{Float64, 1})
pha=Array{Float64}(undef,  grid+1)
    pha[1]=z[1]
i=2
for j=2:grid+1
        if u[j]>d[end]
     pha[j]=z[end]
        else
        while u[j]>d[i]
            i=i+1
end
        pha[j]=z[i-1]+(z[i]-z[i-1])*(u[j]-d[i-1])/(d[i]-d[i-1])
end

end
    return pha
end

function inverse(u, d, z)
pha=Array{Float64}(undef,  grid+1)
    pha[1]=z[1]
i=2
for j=2:grid+1
        if u[j]>d[end]
     pha[j]=z[end]
        else
        while u[j]>d[i]
            i=i+1
end
        pha[j]=z[i-1]+(z[i]-z[i-1])*(u[j]-d[i-1])/(d[i]-d[i-1])
end

end
    return pha
end

#this is a function that, given g: [ucr1, 1]->Cuph, returns (a vector form of) f(x)=int_a^x g(r) dr from ucr1[1] to 1. It uses the trapezoid method to approximate the integral. It's defined a few times to allow for different input types.
function inttrapvec(Cuph::Array{Float64, 1}, ucr1::Float64)
    mass1r1=Array{Float64}(undef,  size(Cuph)[1])
mass1r1[1]=0
for i=2:grid+1
    mass1r1[i]=mass1r1[i-1]+(Cuph[i-1]+Cuph[i])*(1-ucr1)/(2*grid)
end
mass1r1
end
function inttrapvec(Cuph::Array{Float64, 1}, ucr1::Int64)
    mass1r1=Array{Float64}(undef,  size(Cuph)[1])
mass1r1[1]=0
for i=2:grid+1
    mass1r1[i]=mass1r1[i-1]+(Cuph[i-1]+Cuph[i])*(1-ucr1)/(2*grid)
end
mass1r1
end
function inttrapvec(Cuph, ucr1)
    mass1r1=Array{Float64}(undef,  size(Cuph)[1])
mass1r1[1]=0
for i=2:grid+1
    mass1r1[i]=mass1r1[i-1]+(Cuph[i-1]+Cuph[i])*(1-ucr1)/(2*grid)
end
mass1r1
end
#The rescale takes as the input a function f: d->z and then returns a vactor pha, such that u->pha also describes the function f. u is a vector that starts at zero and finishes at 1, with a lenght determined by the parameter grid. This function is needed whenever I need to work with two or more functions that are defined over different vectors of arguments: rescale allows me then to redefine them over vector u. The parameter start is needed whenever d[1]>0 -- it then determines the vleues of f for all arguments which the original specificiation could not define. The value of start usually follows from my paper.
function rescale(d::Array{Float64, 1}, z::Array{Float64, 1}, grid, start=0)    #first argument is the original scale, second the vactor if arguments
    u=collect(0:(1/grid):1)
    pha=Array{Float64}(undef,  grid+1)
    pha[1]=start
i=1
for j=2:grid+1
        if u[j]<d[1]
            pha[j]=start
        else
        if u[j]>d[end]
     pha[j]=z[end]
        else
        while u[j]>d[i]
            i=i+1
end
        pha[j]=z[i-1]+(z[i]-z[i-1])*(u[j]-d[i-1])/(d[i]-d[i-1])
end
        end


end
    return pha
end

function rescale(d, z, grid, start=0)
u=collect(0:(1/grid):1)
    pha=Array{Float64}(undef,  grid+1)
    pha[1]=start
i=1
for j=2:grid+1
        if u[j]<d[1]
            pha[j]=start
        else
        if u[j]>d[end]
     pha[j]=z[end]
        else
        while u[j]>d[i]
            i=i+1
end
        pha[j]=z[i-1]+(z[i]-z[i-1])*(u[j]-d[i-1])/(d[i]-d[i-1])
end
        end
end
    return pha
end

#the following function takes a function x: f and returns its value for an arbitrary argument a such that $a \in (x[1], x[end])$.
#This is done by finding the closes highest and lowest values of x than a, and then linearly approximating the value of f(a).
function makecont(a::Number, x::Array, f::Array)
    i=1
            if a<x[1]
         print("DOMAIN ERROR")
     elseif a>x[end]
         print("DOMAIN ERROR")
     else
            while a>x[i]
                i=i+1
    end
end
if i==1
    return f[1]
else
            return f[i-1]+(f[i]-f[i-1])*(a-x[i-1])/(x[i]-x[i-1])
        end
            end

function makecont(a::Array, x::Array, f::Array)
                i=1
                z=Array{Float64}(undef,  length(a), 1)
                for j=1:length(a)
                        if a[j]<x[1]
                     print("DOMAIN ERROR")
                 elseif a[j]>x[end]
                     print("DOMAIN ERROR")
                 else
                        while a[j]>x[i]
                            i=i+1
                end
            end
            if i==1
                 z[j]=f[1]
            else
                        z[j]=f[i-1]+(f[i]-f[i-1])*(a[j]-x[i-1])/(x[i]-x[i-1])
                    end
                end
return z
                        end
